/*
 *   Copyright (c) 1997-2016 IBM Corporation.  All rights reserved.
 *
 *	Function:	- example: ping-pong benchmark
 *
 *	Usage:		mpirun -np 2 ping_pong [nbytes]
 *
 *	Notes:		- Define CHECK macro to check data integrity.
 *			- The previous ping_pong example timed each
 *			  iteration.  The resolution of MPI_Wtime() is
 *			  not sufficient to provide accurate measurements
 *			  when nbytes is small.  This version times the
 *			  entire run and reports average time to avoid
 *			  this issue.
 *			- To avoid cache conflicts due to direct
 *			  process-to-process bcopy, the buffers are
 *			  aligned and offset from each other.
 *			- Use of direct process-to-process bcopy coupled
 *			  with the fact that the data is never touched
 *			  results in inflated bandwidth numbers when
 *			  nbytes <= cache size.  To obtain a more
 *			  representative bandwidth measurement, increase
 *			  nbytes to at least 2*cache size (2MB).
 *       - For good latency on the same node, try binding
 *         the ranks as close together as possible (not the
 *         default at this time)
 *         --report-bindings --bind-to core --map-by core -prot
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#ifndef _WIN32
#include <unistd.h>
#endif
#include <string.h>
#include <math.h>
#include <mpi.h>

#define NLOOPS		1000
#define ALIGN		4096

#define SEND(t)		MPI_Send(buf, nbytes, MPI_CHAR, partner, (t), \
				MPI_COMM_WORLD)
#define RECV(t)		MPI_Recv(buf, nbytes, MPI_CHAR, partner, (t), \
				MPI_COMM_WORLD, &status)
#ifdef CHECK
#  define SETBUF()	for (j=0; j<nbytes; j++) { \
				buf[j] = (char) (j + i); \
			}
#  define CLRBUF()	memset(buf, 0, nbytes)
#  define CHKBUF()	for (j = 0; j < nbytes; j++) { \
				if (buf[j] != (char) (j + i)) { \
					printf("error: buf[%d] = %d, " \
						"not %d\n", \
						j, buf[j], j + i); \
					break; \
				} \
			}
#else
#  define SETBUF()
#  define CLRBUF()
#  define CHKBUF()
#endif

int
main(int argc, char *argv[])
{
	int		i;
#ifdef CHECK
	int		j;
#endif
	double		start, stop;
	int		nbytes = 0;
	int		rank, size;
	int		root;
	int		partner;
	MPI_Status	status;
	char		*buf, *obuf;
	char		myhost[MPI_MAX_PROCESSOR_NAME];
	int		len;
	char		str[1024];

	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Get_processor_name(myhost, &len);

	if (size < 2) {
		if ( ! rank) printf("rping: must have two+ processes\n");
		MPI_Finalize();
		exit(0);
	}

	nbytes = (argc > 1) ? atoi(argv[1]) : 0;
	if (nbytes < 0) nbytes = 0;
/*
 * Page-align buffers and displace them in the cache to avoid collisions.
 */
	buf = (char *) malloc(nbytes + 524288 + (ALIGN - 1));
	obuf = buf;
	if (buf == 0) {
		MPI_Abort(MPI_COMM_WORLD, MPI_ERR_BUFFER);
		exit(1);
	}

	buf = (char *) ((((MPI_Aint) buf) + (ALIGN - 1)) & ~(ALIGN - 1));
	if (rank > 0) buf += 524288;
	memset(buf, 0, nbytes);
/*
 * Ping-pong.
 */
	for (root=0; root<size; root++) {
		if (rank == root) {
			partner = (root + 1) % size;
			sprintf(str, "[%d:%s] ping-pong %d bytes ...\n",
				root, myhost, nbytes);
/*
 * warm-up loop
 */
			for (i = 0; i < 5; i++) {
				SEND(1);
				RECV(1);
			}
/*
 * timing loop
 */
			start = MPI_Wtime();
			for (i = 0; i < NLOOPS; i++) {
				SETBUF();
				SEND(1000 + i);
				CLRBUF();
				RECV(2000 + i);
				CHKBUF();
			}
			stop = MPI_Wtime();

			sprintf(&str[strlen(str)],
				"%d bytes: %.2f usec/msg\n", nbytes,
				(stop - start) / NLOOPS / 2 * 1024 * 1024);
			if (nbytes > 0) {
				sprintf(&str[strlen(str)],
					"%d bytes: %.2f MB/sec\n", nbytes,
					nbytes / (1024. * 1024.) /
					((stop - start) / NLOOPS / 2));
			}
			fflush(stdout);
		} else if (rank == (root+1)%size) {
/*
 * warm-up loop
 */
			partner = root;
			for (i = 0; i < 5; i++) {
				RECV(1);
				SEND(1);
			}
			for (i = 0; i < NLOOPS; i++) {
				CLRBUF();
				RECV(1000 + i);
				CHKBUF();
				SETBUF();
				SEND(2000 + i);
			}
		}

		MPI_Bcast(str, 1024, MPI_CHAR, root, MPI_COMM_WORLD);
		if (rank == 0) {
			printf("%s", str);
		}
	}

	free(obuf);
	MPI_Finalize();
	exit(0);
}
